import argparse
import inspect
import sys

from . import BasePlugin


class ManagePlugin(BasePlugin):

    name = 'manage'

    def __init__(self, **options):
        super().__init__(**options)

        self.parser = argparse.ArgumentParser(description="Manage Application")
        self.parsers = self.parser.add_subparsers(dest='subparser')
        self.handlers = dict()

    def setup(self, app):

        super().setup(app)

        app.config.setdefault('MANAGE_SHELL', {'app': app})
        self.parser.description = "Manage %s" % app.name.capitalize()

        @self.command
        def shell():
            """ Run the application shell. """
            from werkzeug import script

            ctx = app.config['MANAGE_SHELL']
            if callable(ctx):
                ctx = ctx()

            return script.make_shell(lambda: ctx, "Loaded objects: " + ", ".join(ctx.keys()))()

        @self.command
        def runserver(port=5000):
            """ Run the application. """
            from muffin.worker import GunicornApp
            gapp = GunicornApp(app, {'bind': '127.0.0.1:%d' % port})
            gapp.run()

    def command(self, func):
        parser = self.parsers.add_parser(func.__name__, description=func.__doc__)
        args, varargs, keywords, defaults = inspect.getargspec(func)
        if args:
            for name, value in zip(args, defaults):
                argname = "--%s" % name
                if isinstance(value, bool):
                    parser.add_argument(argname, action="store_%s" % str(not value).lower())
                elif isinstance(value, list):
                    parser.add_argument(argname, action="append", default=value)
                else:
                    parser.add_argument(argname, default=value, type=type(value))
        self.handlers[func.__name__] = func
        return func

    def shell(self, func):
        self.app.config['MANAGE_SHELL'] = func

    def __call__(self, args=None):
        if args is None:
            args = sys.argv[1:]
        options = self.parser.parse_args(args)
        kwargs = dict(options._get_kwargs())
        handler = self.handlers.get(kwargs.pop('subparser'))
        if not handler:
            self.parser.print_help()
            sys.exit(1)

        try:
            handler(**kwargs)
            sys.exit(0)
        except Exception as e:
            sys.exit(e)

# pylama:ignore=E1103,W0612



import base64
import hashlib
import hmac
import time
import asyncio

import ujson as json

from . import BasePlugin


class SessionPlugin(BasePlugin):

    """ Support sessions. """

    name = 'session'
    defaults = {
        'secret': 'InsecureSecret',
    }

    def setup(self, app):
        """ Initialize the application. """
        super().setup(app)

        if self.options['secret'] == 'InsecureSecret':
            app.logger.warn('Use insecure secret key. Change AUTH_SECRET option in configuration.')

    @asyncio.coroutine
    def middleware_factory(self, app, handler):

        @asyncio.coroutine
        def middleware(request):
            request.session = Session(self.options['secret'])
            request.session.load(request.cookies)
            app.logger.debug('Started: %s', request.session)
            response = yield from handler(request)
            app.logger.debug('Ended: %s', request.session)
            request.session.save(response.set_cookie)
            return response
        return middleware


class Session(dict):

    encoding = 'UTF-8'

    def __init__(self, secret, key='session.id', **params):
        self.secret = secret
        self.key = key
        self.params = params
        self.store = {}

    def save(self, set_cookie):
        if set(self.store.items()) ^ set(self.items()):
            value = dict(self.items())
            value = json.dumps(value)
            value = self.encrypt(value)
            if not isinstance(value, str):
                value = value.encode(self.encoding)
            set_cookie(self.key, value, **self.params)

    def load(self, cookies, **kwargs):
        value = cookies.get(self.key, None)
        if value is None:
            return False
        data = json.loads(self.decrypt(value))
        if not isinstance(data, dict):
            return False
        self.store = data
        self.update(self.store)

    def create_signature(self, value, timestamp):
        h = hmac.new(self.secret.encode(), digestmod=hashlib.sha1)
        h.update(timestamp)
        h.update(value)
        return h.hexdigest()

    def encrypt(self, value):
        timestamp = str(int(time.time())).encode()
        value = base64.b64encode(value.encode(self.encoding))
        signature = self.create_signature(value, timestamp)
        return "|".join([value.decode(self.encoding), timestamp.decode(self.encoding), signature])

    def decrypt(self, value):
        value, timestamp, signature = value.split("|")
        check = self.create_signature(value.encode(self.encoding), timestamp.encode())
        if check != signature:
            return None
        return base64.b64decode(value).decode(self.encoding)



import asyncio

from gunicorn.app.base import Application as VanillaGunicornApp
from gunicorn.workers.base import Worker as VanillaGunicornWorker


class GunicornApp(VanillaGunicornApp):

    """ Support Gunicorn. """

    def __init__(self, app, *args, **kwargs):
        super(GunicornApp, self).__init__(*args, **kwargs)
        self.app = app
        self.cfg.set('worker_class', 'muffin.worker.GunicornWorker')
        self.cfg.set('reload', True)

    def load_config(self):
        pass

    def load(self):
        return self.app


class GunicornWorker(VanillaGunicornWorker):

    """ Class description. """

    def __init__(self, *args, **kwargs):
        super(GunicornWorker, self).__init__(*args, **kwargs)
        self.servers = {}

    def init_process(self):
        # create new event_loop after fork
        asyncio.get_event_loop().close()

        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        super().init_process()

    def run(self):
        container = self.app.callable
        container._loop = self.loop
        container.loop.set_debug(container.config['DEBUG'])
        handler = container.make_handler()
        for i, sock in enumerate(self.sockets):
            srv = self.loop.run_until_complete(self.loop.create_server(handler, sock=sock.sock))
            self.servers[srv] = handler

        self.notify()
        self.loop.run_forever()

    def notify(self):
        super(GunicornWorker, self).notify()
        self.loop.call_later(self.timeout, self.notify)



""" Implement Muffin Application. """
import asyncio
import logging
import os
from importlib import import_module

import ujson as json
from aiohttp import web
from cached_property import cached_property


CONFIGURATION_ENVIRON_VARIABLE = 'MUFFIN_CONFIG'


class MuffinException(Exception):

    """ Implement a Muffin's exception. """

    pass


class Plugins(dict):

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)


class Application(web.Application):

    """ Do some stuff. """

    # Default application settings
    __defaults = {

        # Install the plugins
        'PLUGINS': (
            'muffin.plugins.manage:ManagePlugin',
            'muffin.plugins.jade:JadePlugin',
            'muffin.plugins.peewee:PeeweePlugin',
            'muffin.plugins.session:SessionPlugin',
        ),

        # Configuration module
        'CONFIG': 'config',

        # Enable debug mode
        'DEBUG': False
    }

    def __init__(self, name, *, loop=None, router=None, middlewares=(), logger=web.web_logger,
                 handler_factory=web.RequestHandlerFactory, **OPTIONS):
        """ Initialize the application. """
        super().__init__(loop=loop, router=router, middlewares=middlewares, logger=logger,
                         handler_factory=handler_factory)

        self.name = name
        self.defaults = dict(self.__defaults)
        self.defaults.update(OPTIONS)

        self._middlewares = list(self._middlewares)

        # Setup logging
        ch = logging.StreamHandler()
        ch.setFormatter(logging.Formatter(
            '%(asctime)s [%(process)d] [%(levelname)s] %(message)s',
            '[%Y-%m-%d %H:%M:%S %z]'))
        self.logger.addHandler(ch)
        self.logger.setLevel('DEBUG') if self.config['DEBUG'] else self.logger.setLevel('WARNING')

        self.plugins = Plugins()
        for plugin in self.config['PLUGINS']:
            try:
                self.install(plugin)
            except Exception as exc:
                self.logger.error('Plugin is invalid: %s (%s)' % (plugin, exc))

    @cached_property
    def config(self):
        """ Load the application configuration. """
        config = dict(self.defaults)
        module = os.environ.get(CONFIGURATION_ENVIRON_VARIABLE, config['CONFIG'])
        try:
            module = import_module(module)
            config.update({
                name: getattr(module, name) for name in dir(module)
                if name == name.upper() and not name.startswith('_')
            })
        except ImportError:
            self.logger.warn("The configuration hasn't found: %s" % module)

        return config

    def install(self, plugin):
        """ Install plugin to the application. """

        if isinstance(plugin, str):
            module, attr = plugin.split(':')
            module = import_module(module)
            plugin = getattr(module, attr)

        if isinstance(plugin, type):
            plugin = plugin()

        if hasattr(plugin, 'setup'):
            plugin.setup(self)

        if hasattr(plugin, 'middleware_factory') \
                and plugin.middleware_factory not in self.middlewares:
            self.middlewares.append(plugin.middleware_factory)

        self.plugins[plugin.name] = plugin

    def view(self, path, method='GET', name=None):
        """ Convert a view to couroutine and bind a route. """
        def wrapper(view):
            if not asyncio.iscoroutinefunction(view):
                view = asyncio.coroutine(view)

            def wrap_response(request, **kwargs):
                response = yield from view(request, **kwargs)

                if isinstance(response, (list, dict)):
                    response = web.Response(text=json.dumps(response),
                                            content_type='application/json')

                elif not isinstance(response, web.Response):
                    response = web.Response(text=str(response), content_type='text/html')

                self.logger.info('%s %d %s', request.method, response.status, request.path)
                return response

            self.router.add_route(method, path, wrap_response, name=name)

        return wrapper



from . import BasePlugin
import flask as f
import flask_admin as fa


class AdminPlugin(BasePlugin):

    """ Implement admin interface. """

    name = 'admin'
    defaults = {
        'route': '/admin',
    }

    def setup(self, app):
        super().setup(app)

        fapp = f.Flask(app.name)
        self.admin = fa.Admin(fapp, url=self.options['route'])
